#!perl
use strict;
use warnings;
use utf8;

use List::Util qw/uniq/;
use List::UtilsBy qw/uniq_by/;
use PPI;
use PPI::Find;
use PPI::Dumper;
use PPR;
use Path::Tiny;
use Perl::Tidy;
use Text::WordDiff qw/word_diff/;
use Scalar::Util qw/blessed/;
use Cwd qw/getcwd/;

my $caller_dir = shift @ARGV;
my @files      = @ARGV;
my $dir_p      = path($caller_dir);
while (!$dir_p->child(".git")->is_dir) {
    $dir_p = $dir_p->parent;
    if ($dir_p->stringify eq "/") {
        die ".git is not found";
    }
}

use Git::Repository;
my $repo = Git::Repository->new(work_tree => $dir_p->stringify);

my @diff_outputs = $repo->run(
    qw/diff --diff-filter M -U0 *.pm *.pl/
);

my %file_lines;
my $filename_regexp = qr!
    \+\+\+\sb/(.*)$
!x;
my $line_regexp = qr/
    ^@@\s-(?<minus_line_n>\d+)(?:,(?<minus_line_c>\d+))?\s\+(?<plus_line_n>\d+)(?:,(?<plus_line_c>\d+))?\s@@
/x;
my $current_filename;
for my $diff_line (@diff_outputs) {
    if (my ($filename) = $diff_line =~ $filename_regexp) {
        $current_filename = $filename;
    }
    if ($diff_line !~ $line_regexp) {
        next;
    }
    my ($minus_line_n, $plus_line_n) = @+{qw/minus_line_n plus_line_n/};
    my ($minus_line_c, $plus_line_c) = ($+{minus_line_c} // 1, $+{plus_line_c} // 1);
    if ($minus_line_c > 0) {
        push @{ $file_lines{$current_filename} //= [] }, map { $minus_line_n + $_ } 0 .. ($minus_line_c - 1);
    }
    if ($plus_line_c > 0) {
        push @{ $file_lines{$current_filename} //= [] }, map { $plus_line_n + $_ } 0 .. ($plus_line_c - 1);
    }
}

for my $filename (keys %file_lines) {
    $file_lines{$filename} = [sort { $a <=> $b } uniq @{ $file_lines{$filename} }];
}

for my $filename (keys %file_lines) {
    if (scalar(@files) > 0 && !grep { $filename eq $_ } @files) {
        next;
    }

    my $line_numbers = $file_lines{$filename};
    my $path         = $dir_p->child($filename);
    my $fullpath     = $path->stringify;
    # utf8文字化け対策のため一旦文字列で読んでいます
    my $filetext = $path->slurp_utf8;
    my $doc      = PPI::Document->new(\$filetext);
    my $nodes    = $doc->find(
        sub {
            my (undef, $node) = @_;
            return grep { $node->line_number == $_ } @$line_numbers;
        }
    );
    my @matches_statement;
    for my $node (@$nodes) {
        my $current_node = $node;
        while (!$current_node->parent->scope) {
            $current_node = $current_node->parent;
        }
        next if $current_node->isa("PPI::Token");
        if ($current_node->parent->isa("PPI::Statement::Compound")) {
            $current_node = $current_node->parent;
        }
        push @matches_statement, $current_node;
    }
    @matches_statement = uniq_by { $_->line_number . "-" . $_->column_number } @matches_statement;
    @matches_statement = sort {
        $a->line_number <=> $b->line_number ||
            $a->column_number <=> $b->column_number
    } @matches_statement;
    my @reduced_statement;
    for my $stmt (@matches_statement) {
        my $prev = scalar(@reduced_statement) > 0 ? $reduced_statement[-1] : undef;
        next if $prev && $prev->contains($stmt);
        push @reduced_statement, $stmt;
    }
    # ループ内で生成したPPI::Documentのスコープが外れると結果がおかしくなるため
    # この配列に収納することで消えることを防止しています
    my @guards;
    for my $stmt (@reduced_statement) {
        my $indent_content = "";
        my $before_indent  = $stmt->previous_sibling;
        if (blessed $before_indent && $before_indent->isa("PPI::Token::Whitespace")) {
            $indent_content = $before_indent->content;
            $indent_content =~ s/\A\n//gs;
        }
        my $content = $stmt->content;
        $content =~ s/^/$indent_content/g;
        my $dest_content = "";
        my $stderr       = "";

        my $err = Perl::Tidy::perltidy(
            argv        => "",
            source      => \$content,
            destination => \$dest_content,
            stderr      => \$stderr,
        );
        $err and die $err;
        my $dest_pdoc = PPI::Document->new(\$dest_content);
        push @guards, $dest_pdoc;

        my @dest_elements = $dest_pdoc->elements;
        shift @dest_elements if $dest_elements[0]->isa("PPI::Token::Whitespace");
        $stmt->insert_before(@dest_elements);
        $stmt->remove;
    }
    $path->spew_utf8($doc->content);
}
